package us.rddt.IRCBot;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.util.Date;
import java.util.ArrayList;
import java.util.Set;

import org.apache.commons.lang3.exception.ExceptionUtils;
import org.apache.commons.lang3.StringEscapeUtils;
import org.pircbotx.Channel;

/**
 * Utilities and helper methods that are called in various functions of the bot.
 * 
 * @author Ryan Morrison
 */
public class IRCUtils {
    /* 
     * Sun property pointing the main class and its arguments. 
     * Might not be defined on non Hotspot VM implementations.
     */
    public static final String SUN_JAVA_COMMAND = "sun.java.command";

    /**
     * Returns a properly escaped string value
     * @param toEscape the string to escape
     * @return the escaped string
     */
    public static String escapeHTMLEntities(String toEscape) {
        new StringEscapeUtils();
        return StringEscapeUtils.unescapeHtml4(toEscape);
    }
    
    /**
     * Returns an error string appropriate for an HTTP status code.
     * @param statusCode the HTTP status code
     * @return the error string
     */
    public static String getHttpStatusErrorString(int statusCode) {
    	switch(statusCode) {
    		case 400:
    			return "The request was syntatically invalid.";
    		case 401:
    			return "The request requires user authentication which was not provided.";
    		case 403:
    			return "The request was forbidden.";
    		case 404:
    			return "The request failed as the resource could not be found.";
    		case 405:
    			return "The method specified is not suitable for the request.";
    		case 406:
    			return "The server will not return in a format deemed acceptable in the client request.";
    		case 408:
    			return "The server did not receive a proper request during its timeout period.";
    		case 410:
    			return "The resource no longer exists.";
    		case 500:
    			return "The server experienced an internal server error and is incapable of processing the request.";
    		case 501:
    			return "The server does not implement the functionality required to process the request.";
    		case 502:
    			return "The server received an invalid response from a upstream provider.";
    		case 503:
    			return "The server is temporarily unavailable due to overloading or maintenance.";
    		case 504:
    			return "The server did not receive a response from a upstream provider in time.";
    		case 505:
    			return "The server does not support the HTTP version specified in the request.";
    		default:
    			return "The server could not process the request for an unknown reason.";	
    	}
    }
    
    /**
     * Returns a more user-friendly string containing the cipher suite of a HTTPS connection.
     * Java tends to return strings such as TLS_ECDHE_RSA_WITH_RC4_128_SHA, when we only really need to
     * show the user it's a 128-bit secured TLS connection.
     * @see java.net.HttpsURLConnection
     * @param cipherSuite the output of HttpsURLConnections' getCipherSuite method
     * @return the formatted cipher suite to display to the user
     */
    public static String getReadableCipherSuite(String cipherSuite) {
        StringBuilder readableSuite = new StringBuilder();
        // The first element in a split cipher string always defines a TLS or SSL connection
        readableSuite.append(cipherSuite.split("_")[0] + " ");
        // Determine the bit strength
        if(cipherSuite.contains("NULL")) readableSuite.append("0-bit null cipher");
        if(cipherSuite.contains("40")) readableSuite.append("40-bit");
        if(cipherSuite.contains("128")) readableSuite.append("128-bit");
        if(cipherSuite.contains("3DES")) readableSuite.append("162-bit");
        if(cipherSuite.contains("256")) readableSuite.append("256-bit");
        // Return our nicely formatted string
        return readableSuite.toString();
    }
    
    /**
     * Returns the stack trace of a throwable object as a string
     * @param throwable the throwable object to return the stack trace for
     * @return the stack trace as a string
     */
    public static String getStackTraceString(Throwable throwable) {
        new ExceptionUtils();
        return ExceptionUtils.getStackTrace(throwable);
    }
    
    /**
     * Restart the current Java application
     * @param channels the channels to join to when the bot restarts
     * @throws URISyntaxException 
     */
    public static void restartApplication(Set<Channel> channels) throws IOException, URISyntaxException
    {
      final String javaBin = System.getProperty("java.home") + File.separator + "bin" + File.separator + "java";
      final File currentJar = new File(IRCBot.class.getProtectionDomain().getCodeSource().getLocation().toURI());

      /* is it a jar file? */
      if(!currentJar.getName().endsWith(".jar"))
        return;

      /* Build command: java -jar application.jar */
      final ArrayList<String> command = new ArrayList<String>();
      command.add(javaBin);
      command.add("-jar");
      command.add(currentJar.getPath());
      for(Channel c : channels) {
          command.add(c.getName());
      }

      final ProcessBuilder builder = new ProcessBuilder(command);
      builder.start();
      System.exit(0);
    }

    /**
     * Converts a time format in long form to a human-readable minute value
     * @param time the time to be converted
     * @return the human-readable minute value
     */
    public static String toReadableMinutes(long time) {
        if(time > 3600) {
            return String.format("%d:%02d:%02d", time/3600, (time % 3600)/60, (time % 60));
        } else {
            return String.format("%d:%02d", (time % 3600)/60, (time % 60));
        }
    }

    /**
     * Converts date formats to human-readable strings
     * @param date the date to be converted
     * @param countingDown true if we are counting down to a date, false if we are determining history
     * @param showMoment true to show seconds as "a moment" and one hour as "an", false to show the number of seconds and hours
     * @return the human-readable string
     */
    public static String toReadableTime(Date date, boolean countingDown, boolean showMoment) {
        // Calculate the difference in seconds between the time the user left and now
        long diffInSeconds;
        if(countingDown) diffInSeconds = (date.getTime() - System.currentTimeMillis()) / 1000;
        else diffInSeconds = (System.currentTimeMillis() - date.getTime()) / 1000;

        // Calculate the appropriate day/hour/minute/seconds ago values and insert them into a long array
        long diff[] = new long[] { 0, 0, 0, 0 };
        diff[3] = (diffInSeconds >= 60 ? diffInSeconds % 60 : diffInSeconds);
        diff[2] = (diffInSeconds = (diffInSeconds / 60)) >= 60 ? diffInSeconds % 60 : diffInSeconds;
        diff[1] = (diffInSeconds = (diffInSeconds / 60)) >= 24 ? diffInSeconds % 24 : diffInSeconds;
        diff[0] = (diffInSeconds = (diffInSeconds / 24));

        // Build the readable format string
        if(diff[0] != 0) return String.format("%d day%s", diff[0], diff[0] > 1 ? "s" : "");
        if(diff[1] != 0) {
            if(showMoment) return String.format("%s%s hour%s", diff[1] > 1 ? "" : "an", diff[1] > 1 ? String.valueOf(diff[1]) : "", diff[1] > 1 ? "s" : "");
            else return String.format("%s hour%s", diff[1], diff[1] > 1 ? "s" : "");
        }
        if(diff[2] != 0) return String.format("%d minute%s", diff[2], diff[2] > 1 ? "s" : "");
        if(diff[3] != 0) {
            if(showMoment) return "a moment";
            else return String.format("%d second%s", diff[3], diff[3] > 1 ? "s" : "");
        }
        else return "unknown";
    }

    /**
     * Trims strings greater than a provided length
     * @param toTrim the string to be trimmed, if necessary
     * @param maxLength the maximum length the string can be
     * @return the trimmed string
     */
    public static String trimString(String toTrim, int maxLength) {
        if(toTrim.length() > maxLength) {
            return toTrim.substring(0, maxLength) + "...";
        } else {
            return toTrim;
        }
    }
}
